local function SafeFreeSlot(...)
	for _,slot in ipairs({...})
		if not rawget(_G, slot) freeslot(slot) end -- Ensure we don't accidentally overlap existing freeslots
	end
end
SafeFreeSlot(
-- States
"S_WARSPORT","S_WARMPORT","S_WARLPORT","S_HURRYUP","S_COINLOSS",
-- Objects
"MT_WLPORTALSMALL","MT_WLPORTALMEDIUM","MT_WLPORTALLARGE","MT_FROGSWITCH","MT_COINLOSSEFFECT",
-- Sprites
"SPR_WLPS","SPR_WLPM","SPR_WLPL","SPR_KERO","SPR_WLC2",
-- Wario's Voice (Hurry Up!)
"sfx_hurry0","sfx_hurry1","sfx_hurry2","sfx_hurry3","sfx_hurry4",
-- Kero SFX
"sfx_wlohno","sfx_wlclos")

local hurryup = false
local mapclock =  180*TICRATE
local hittime = 0
local outtatimegong = false
local coinxoff = 0
local coinsize = FRACUNIT

sfxinfo[sfx_hurry0].caption = '"Hurry up!"'
sfxinfo[sfx_hurry1].caption = '"Hurry up!"'
sfxinfo[sfx_hurry2].caption = '"Hurry up!"'
sfxinfo[sfx_hurry3].caption = '"H-H-H-HURRY UP!"'
sfxinfo[sfx_hurry4].caption = '"No, no, no! Grrr... Hurry up!"'
sfxinfo[sfx_wlohno].caption = 'Only a few seconds left!!!'
sfxinfo[sfx_wlclos].caption = 'Coin loss...'
sfxinfo[sfx_wlohno].singular = true
sfxinfo[sfx_wlclos].singular = true

local inescapable = {
	["deep sea zone 2"] = true,
	["castle eggman zone 1"] = true,
	["red volcano zone 1"] = true,
	["egg rock zone 1"] = true,
	["black core zone 1"] = true,
	["pipe towers zone"] = true,
	["hell coaster zone 1"] = true,
	["festung oder so"] = true,
	["spiral hill pizza"] = true
}

rawset(_G, "cv_kombiforceescape", CV_RegisterVar({
	name = "k_forcehurryup",
	defaultvalue = "Off",
	flags = CV_SAVE|CV_SHOWMODIF|CV_NETVAR,
	PossibleValue = {On = 1, Off = 0},
}))

rawset(_G, "cv_kombinoescape", CV_RegisterVar({
	name = "k_nohurryup",
	defaultvalue = "Off",
	flags = CV_SAVE|CV_SHOWMODIF|CV_NETVAR,
	PossibleValue = {On = 1, Off = 0},
}))

rawset(_G, "cv_kombinetescape", CV_RegisterVar({
	name = "k_netgamekero",
	defaultvalue = "Off",
	flags = CV_SAVE|CV_SHOWMODIF|CV_NETVAR,
	PossibleValue = {On = 1, Off = 0},
}))

COM_AddCommand("k_exitlevel", function(player)
	if gamestate ~= GS_LEVEL
		print("You need a level to exit.")
		return
	end
	if hurryup == false
		print("Can't do this right now.")
		return
	end
	P_DoPlayerFinish(player)
end)

addHook("MapLoad", function(mapid)
	hurryup = false
	hittime = 0
	coinxoff = 0
	coinsize = FRACUNIT
	local haskombi = false
	for player in players.iterate
		player.exitcutscene = 0
		P_RestoreMusic(player)
		if skins[player.skin].name == "kombi"
			haskombi = true
		end
	end
	if cv_kombinoescape.string == "On" then return end
	if not (haskombi and CV_FindVar("k_forcehurryup").string == "Off") then return end
	if (inescapable[string.lower(G_BuildMapTitle(gamemap))]) == true and CV_FindVar("k_forcehurryup").string == "Off" then return end
	if (netgame) and CV_FindVar("k_netgamekero").string == "Off" then return end
	if (G_IsSpecialStage(mapid))
	or (maptol & TOL_NIGHTS)
	or (mapheaderinfo[gamemap].bonustype == 1)
	or (gamemap == titlemap)
	or G_BuildMapTitle(gamemap) == nil then return end
	
	local hassign = false
	local hasportal = false
	
	for mthing in mapthings.iterate
		if mthing.type == 1
			if not hasportal
				local smallportal = P_SpawnMobj(mthing.x*FRACUNIT,mthing.y*FRACUNIT,mthing.z*FRACUNIT,MT_WLPORTALSMALL)
				local mediumportal = P_SpawnMobj(mthing.x*FRACUNIT,mthing.y*FRACUNIT,mthing.z*FRACUNIT,MT_WLPORTALMEDIUM)
				local largeportal = P_SpawnMobj(mthing.x*FRACUNIT,mthing.y*FRACUNIT,mthing.z*FRACUNIT,MT_WLPORTALLARGE)
				hasportal = true
			end
		end
		if mthing.type == 501
			local trig = P_SpawnMobjFromMobj(mthing.mobj,0,0,0,MT_FROGSWITCH)
			if (mthing.mobj and mthing.mobj.valid) then P_RemoveMobj(mthing.mobj) end
			hassign = true
		end	
		if mthing.type == 312
			local keyzer = P_SpawnMobjFromMobj(mthing.mobj,0,0,0,MT_KEYZER)
		end	
	end
	
	local SSF_REALEXIT = 1<<7
	if hassign
		for sec in sectors.iterate
			if (sec.specialflags & SSF_REALEXIT)
			or (GetSecSpecial(sec.special, 4) == 2)
				sec.specialflags = $ &~SSF_REALEXIT
			end
		end
	end
end)

addHook("MobjThinker", function(mobj)
	if hurryup
		P_RemoveMobj(mobj) -- There is no such thing as a Metal Sonic.
	end
end, MT_METALSONIC_RACE)

mobjinfo[MT_WLPORTALSMALL] = {
spawnstate = S_WARSPORT,
spawnhealth = 1000,
deathstate = S_WARSPORT,
radius = mobjinfo[MT_PLAYER].radius,
height = mobjinfo[MT_PLAYER].height,
dispoffset = 5,
flags = MF_SCENERY|MF_NOGRAVITY|MF_NOCLIPHEIGHT,
}

mobjinfo[MT_WLPORTALMEDIUM] = {
spawnstate = S_WARMPORT,
spawnhealth = 1000,
deathstate = S_WARMPORT,
radius = mobjinfo[MT_PLAYER].radius,
height = mobjinfo[MT_PLAYER].height,
dispoffset = 4,
flags = MF_SCENERY|MF_NOGRAVITY|MF_NOCLIPHEIGHT,
}

mobjinfo[MT_WLPORTALLARGE] = {
spawnstate = S_WARLPORT,
spawnhealth = 1000,
deathstate = S_WARLPORT,
radius = 24*FRACUNIT,
height = 48*FRACUNIT,
dispoffset = 3,
flags = MF_SCENERY|MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_SPECIAL,
}

mobjinfo[MT_FROGSWITCH] = {
spawnstate = S_HURRYUP,
spawnhealth = 1000,
deathstate = S_HURRYUP,
radius = 12*FRACUNIT,
height = 36*FRACUNIT,
dispoffset = 3,
flags = MF_SCENERY|MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_SOLID,
}

mobjinfo[MT_COINLOSSEFFECT] = {
spawnstate = S_COINLOSS,
spawnhealth = 1000,
deathstate = S_COINLOSS,
radius = mobjinfo[MT_PLAYER].radius,
height = mobjinfo[MT_PLAYER].height,
dispoffset = 4,
flags = MF_SCENERY|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING, -- i actually don't know if we need NOCLIPTHING
}

states[S_WARSPORT] = {
	sprite = SPR_WLPS,
	frame = FF_ANIMATE|A,
	tics = -1,
	var1 = 32,
	var2 = 10,
	nextstate = S_WARSPORT,
}

states[S_WARMPORT] = {
	sprite = SPR_WLPM,
	frame = FF_ANIMATE|A,
	tics = -1,
	var1 = 32,
	var2 = 10,
	nextstate = S_WARMPORT,
}

states[S_COINLOSS] = {
	sprite = SPR_WLC2,
	frame = FF_ANIMATE|A,
	tics = 36,
	var1 = 5,
	var2 = 2,
	nextstate = S_NULL,
}

states[S_WARLPORT] = {
	sprite = SPR_WLPL,
	frame = FF_ANIMATE|A,
	tics = -1,
	var1 = 32,
	var2 = 10,
	nextstate = S_WARLPORT,
}

states[S_HURRYUP] = {
	sprite = SPR_KERO,
	tics = -1,
	var1 = 0,
	var2 = 0,
	nextstate = S_HURRYUP,
}

local srbtpmetaltimers = {
	1958,
	2396,
	2133,
	3253,
	3007,
	2203,
	3113,
	2448,
	3761,
	3428,
	5843,
	2133,
	7208,
	3708,
	2712,
	2308,
	2712,
	2308,
	5231,
	10449,
	3707,
	4872,
	18476
}

local maptimers = {
	["greenflower zone 1"] = 50,
	["greenflower zone 2"] = 50,
	["techno hill zone 1"] = 70,
	["techno hill zone 2"] = 130,
	["deep sea zone 1"] = 105,
	["black core zone 1"] = 110,
	["frozen hillside zone"] = 50,
	["haunted heights zone"] = 120,
	["neo palmtree zone"] = 180
}

addHook("PlayerSpawn", function(player)
	player.timeleft = 20965
	if player.realtime >= 3*TICRATE
		player.disptime = (player.timeleft-(player.realtime-3*TICRATE))
	else
		player.disptime = player.timeleft
	end
	player.timeleftmins = player.disptime/(TICRATE*60)
	player.timeleftsecs =  (player.disptime/TICRATE)%60
	player.timeleftcents = (((player.disptime*100)%3500)/35)
	mapclock = ((maptimers[string.lower(G_BuildMapTitle(gamemap))] or mapheaderinfo[gamemap].takis_hh_timelimit) or 90)*TICRATE -- reusing other people's tags... banished to the shadow realm
	outtatimegong = false
	player.recordscore = $ or 0
end)

addHook("PlayerThink", function(player)
	if gamemap >= 604 and gamemap <= 624
		player.timeleft = srbtpmetaltimers[gamemap - 603]
	end
	if gamemap == 25 and mapheaderinfo[gamemap].lvlttl == "Black Core"
		player.timeleft = 3340
	end
	if hurryup
		player.timeleft = (mapclock + hittime)
	end
	if player.realtime >= 3*TICRATE or hittime > 0
		player.disptime = (player.timeleft-(player.realtime-3*TICRATE))
	else
		player.disptime = player.timeleft
	end
	if player.disptime == nil
		player.disptime = 0
	end
	if player.disptime < 0 and hurryup
		if (player.realtime-hittime)%6 == 0 and hurryup
			local coinloss = P_SpawnMobjFromMobj(player.mo, 0, 0, mobjinfo[MT_PLAYER].height, MT_COINLOSSEFFECT)
			coinloss.angle = player.drawangle
			P_InstaThrust(coinloss, coinloss.angle, -4*FRACUNIT)
			P_SetObjectMomZ(coinloss, 10*FRACUNIT, false)
		end
		if not outtatimegong
			S_ChangeMusic("wlclos", true)
			S_StartSound(nil, sfx_wlclos)
			outtatimegong = true
		end
		if player.disptime%4 == 0 and player.recordscore > 0 and not (player.pflags & PF_FINISHED)
			P_AddPlayerScore(player, -10)
			if player.recordscore == 0 and player.mo.health
				P_KillMobj(player.mo, nil, nil, DMG_SPACEDROWN)
			end
		else
			if player.recordscore == 0 and player.mo.health
				P_KillMobj(player.mo, nil, nil, DMG_SPACEDROWN)
			end
		end
		player.disptime = 0
	end
	player.timeleftmins = player.disptime/(TICRATE*60)
	player.timeleftsecs =  (player.disptime/TICRATE)%60
	player.timeleftcents = (((player.disptime*100)%3500)/35)
	if player.disptime == 385 and hurryup
		S_StartSound(nil, sfx_wlohno)
	end
end)

if not srb2p
	hud.add(function(v,player)
		local offset = 16
		local center = 160-(offset*1)
		local wltimeclr = 2
		if gamemap >= 604 and gamemap <= 624 or (gamemap == 25 and mapheaderinfo[gamemap].lvlttl == "Black Core") or hurryup
			if player.disptime <= 11*TICRATE
				wltimeclr = SKINCOLOR_RED
			elseif player.disptime <= 31*TICRATE
				wltimeclr = SKINCOLOR_YELLOW
			else
				wltimeclr = SKINCOLOR_WHITE
			end
			if not outtatimegong
			if (player.realtime-hittime)%35 < 18
				v.drawScaled((center+offset)*FRACUNIT, hudinfo[HUD_SCORE].y*FRACUNIT, 2*FRACUNIT, v.cachePatch("WLTIMEC"),V_SNAPTOTOP|V_PERPLAYER,v.getColormap(nil, wltimeclr))
			end
			local digit = v.cachePatch("WLTIME\$player.timeleftsecs%10\")
			v.drawScaled((center+(offset*3))*FRACUNIT, hudinfo[HUD_SCORE].y*FRACUNIT, 2*FRACUNIT, digit,V_SNAPTOTOP|V_PERPLAYER,v.getColormap(nil, wltimeclr))
			local digit = v.cachePatch("WLTIME\$player.timeleftsecs/10\")
			v.drawScaled((center+(offset*2))*FRACUNIT, hudinfo[HUD_SCORE].y*FRACUNIT, 2*FRACUNIT, digit,V_SNAPTOTOP|V_PERPLAYER,v.getColormap(nil, wltimeclr))
			local digit = v.cachePatch("WLTIME\$player.timeleftmins\")
			v.drawScaled(center*FRACUNIT, hudinfo[HUD_SCORE].y*FRACUNIT, 2*FRACUNIT, digit,V_SNAPTOTOP|V_PERPLAYER,v.getColormap(nil, wltimeclr))
			end
			local frame =((player.realtime-hittime)/3)%6
			local digit = v.cachePatch("WLCLOCK\$frame\")
			v.draw(center-16, hudinfo[HUD_SCORE].y, digit,V_SNAPTOTOP|V_PERPLAYER)
		end
		v.drawScaled(((320*FRACUNIT)+coinxoff-(((offset/2)*coinsize)*8)-coinsize), (hudinfo[HUD_SCORE].y+2)*FRACUNIT, coinsize, v.cachePatch("WLCOINC"), V_SNAPTOTOP|V_PERPLAYER)
		for i = 0,5 do
			local frame = player.recordscore/(10^i)%10 or 0
			local digit = v.cachePatch("WLCOIN\$frame\")
			v.drawScaled((320*FRACUNIT)+coinxoff-(((offset/2)*coinsize)*(2+i)), (hudinfo[HUD_SCORE].y+2)*FRACUNIT, coinsize, digit, V_SNAPTOTOP|V_PERPLAYER)
		end
		if outtatimegong
			if coinxoff == -46*FRACUNIT then return end
			if coinxoff == -96*FRACUNIT
				coinxoff = -46*FRACUNIT
				coinsize = 2*FRACUNIT
				return
			end
			coinxoff = $ - 12*FRACUNIT
		end
	end,"game")
end

addHook("MobjThinker", function(mobj)
	mobj.rollangle = $ + FixedDiv(ANGLE_45, 45*FRACUNIT)
	if hurryup
		mobj.destscale = 2*FRACUNIT
		mobj.scalespeed = FixedDiv(FRACUNIT, 24*FRACUNIT)
	else
		mobj.scale = FixedDiv(FRACUNIT, 4*FRACUNIT)
	end
end, MT_WLPORTALSMALL)

addHook("MobjThinker", function(mobj)
	mobj.rollangle = $ + FixedDiv(ANGLE_45, 30*FRACUNIT)
	if hurryup
		mobj.destscale = 2*FRACUNIT
		mobj.scalespeed = FixedDiv(FRACUNIT, 24*FRACUNIT)
	else
		mobj.scale = FixedDiv(FRACUNIT, 4*FRACUNIT)
	end
end, MT_WLPORTALMEDIUM)

addHook("MobjThinker", function(mobj)
	mobj.rollangle = $ + FixedDiv(ANGLE_45, 25*FRACUNIT)
	if hurryup
		mobj.destscale = 2*FRACUNIT
		mobj.scalespeed = FixedDiv(FRACUNIT, 24*FRACUNIT)
	else
		mobj.scale = FixedDiv(FRACUNIT, 4*FRACUNIT)
	end
end, MT_WLPORTALLARGE)

addHook("MobjThinker", function(mobj)
	mobj.scale = 2*FRACUNIT
end, MT_FROGSWITCH)

local function PortalSpawn(mobj)
	P_SetOrigin(mobj, mobj.x, mobj.y, mobj.z+mobj.subsector.sector.floorheight+(128*FRACUNIT))
end
addHook("MobjSpawn", PortalSpawn, MT_WLPORTALSMALL)
addHook("MobjSpawn", PortalSpawn, MT_WLPORTALMEDIUM)
addHook("MobjSpawn", PortalSpawn, MT_WLPORTALLARGE)

addHook("MobjCollide",function(switch,mo)
	if hurryup then return end
	if not mo or not mo.valid
		return
	end
	if (mo.type ~= MT_PLAYER) then return end
	if mo.z - switch.z == 74*FRACUNIT
		if mo.momz == 0
			S_ChangeMusic("hurryu", true)
			S_StartSound(mo, P_RandomRange(sfx_hurry0,sfx_hurry4))
			hurryup = true
			hittime = mo.player.realtime-(2*TICRATE)
		end
	end
end,MT_FROGSWITCH)

addHook("TouchSpecial",function(port,mo)
	if hurryup and not mo.player.exiting
		mo.player.exitcutscene = 1
		P_DoPlayerFinish(mo.player)
		mo.momx = 0
		mo.momy = 0
	end
	return true
end,MT_WLPORTALLARGE)

local priority = {
	["_minv"] = true,
	["_inv"] = true,
	["_shoes"] = true,
	["wlclos"] = true,
	["_drown"] = true,
	["sacler"] = true,
	["clear"] = true
}

addHook("MusicChange", function(oldname, newname, mflags, looping, position, prefadems, fadeinms)
	if newname == "_title" and CV_FindVar("k_wl4title").string == "On"
		return "_wl4ti", mflags, looping, position, 0, fadeinms -- make a different music file to ensure we DON'T clash with _TITLE if we want to disable the modified title screen... But why would you?
	end
	if not (consoleplayer and consoleplayer.valid)
		return
	end
	if newname == "sacler"
		hurryup = false
	end
	if hurryup and not priority[newname] then return "hurryu" end
	if (skins[consoleplayer.skin].name != "kombi") or not skins[consoleplayer.skin]
		return
	end
	if CV_FindVar("k_jinglereplacements").string == "Off" or (oldname == "opnyht" or oldname == "livean" or oldname == "sheroes" and consoleplayer.powers[pw_super]) or not (consoleplayer and consoleplayer.valid)
		if newname == "_super" -- redundancy my behated
			if CV_FindVar("k_supertheme").string == "SA1"
				return "opnyht", mflags, looping, position, 0, fadeinms
			elseif CV_FindVar("k_supertheme").string == "SA2"
				return "livean", mflags, looping, position, 0, fadeinms
			elseif CV_FindVar("k_supertheme").string == "SH"
				return "sheroes", mflags, looping, position, 0, fadeinms
			end
		end
		return
	end
	
	if newname == "_super"
		if CV_FindVar("k_supertheme").string == "SA1"
			return "opnyht", mflags, looping, position, 0, fadeinms
		elseif CV_FindVar("k_supertheme").string == "SA2"
			return "livean", mflags, looping, position, 0, fadeinms
		elseif CV_FindVar("k_supertheme").string == "SH"
			return "sheroes", mflags, looping, position, 0, fadeinms
		end
	end

	if not consoleplayer.powers[pw_super]
		if CV_FindVar("k_jinglereplacements").value == 1
	if newname == "_shoes"	
		return "heyyou", mflags, looping, position, 0, fadeinms
		elseif newname == "_inv" or newname == "_minv"
			return "nofear", mflags, looping, position, 0, fadeinms
		end
		elseif CV_FindVar("k_jinglereplacements").string == "SA2"
			if newname == "_shoes"	
				return "sagain", mflags, looping, position, 0, fadeinms
			elseif newname == "_inv" or newname == "_minv"
				return "stilli", mflags, looping, position, 0, fadeinms
			end
		end
	end

	if newname == "_clear" or newname == "_CLEAR"
		return "sacler"
	end
	
	if newname == "_drown"
		return "sadrwn", mflags, looping, position, 0, fadeinms
	end
	
	if CV_FindVar("k_bossreplacements").value == 1
		
	end
end)