local function SafeFreeSlot(...)
	for _,slot in ipairs({...})
		if not rawget(_G, slot) freeslot(slot) end -- Ensure we don't accidentally overlap existing freeslots
	end
end
SafeFreeSlot("sfx_sready", "sfx_slspgo", "sfx_salspd", "sfx_lspdst", "sfx_lspdlo")
sfxinfo[sfx_sready].caption = '\x85"READYYY!"\x80'
sfxinfo[sfx_slspgo].caption = '\x85"GO!!"\x80'
sfxinfo[sfx_salspd].caption = 'Lightspeed Dash'
sfxinfo[sfx_lspdst].caption = 'Lightspeed Dash Starting'
sfxinfo[sfx_lspdlo].caption = ''


rawset(_G, "cv_kombilspd", CV_RegisterVar({
	name = "k_lightspeeddash",
	defaultvalue = "On",
	flags = CV_SAVE|CV_SHOWMODIF,
	PossibleValue = {On = 1, Off = 0},
	category = "Kombi - Gameplay",
	displayname = "Light-Speed Dash",
}))


addHook("PlayerSpawn", function(player)
	if (player.mo.skin == "kombi") and not srb2p
		player.lightspeedhold = 0
		player.lightspeeddashable = 0
		player.timervar2 = 0
	end
end)


addHook("SpinSpecial", function(player)
	if (player.mo.skin == "kombi") and not srb2p
	    and (P_IsObjectOnGround(player.mo))
		and not (player.mo.state == S_PLAY_PAIN)
		and not (player.mo.state == S_PLAY_DEAD)
		and not (player.mo.state == S_PLAY_DRWN)
		and not (player.mo.state == S_PLAY_SPINDASH)
		if player.lightspeeddashable > 0
			if CV_FindVar("k_lightspeeddash").string == "On"
				return true
			end
		end
		player.pflags = $|PF_STARTDASH
		player.mo.state = S_PLAY_SPINDASH
		player.drawangle = player.mo.angle
		player.lightspeedhold = leveltime
		S_StartSound(player.mo, sfx_spndsh)
		if CV_FindVar("k_lightspeeddash").string == "On"
			S_StartSound(player.mo, sfx_salspd)
		end
	end	
	return true
end)


addHook("PlayerThink", function(player)
	if player.lightspeedhold == nil
		player.lightspeedhold = 99999*TICRATE
	end
	if player.lightspeeddashable == nil
		player.lightspeeddashable = 0
	end
	if player.timervar2 == nil
		player.timervar2 = 99999*TICRATE
	end
	if (player.lightspeedhold - leveltime) <= -100 and player.mo.state == S_PLAY_SPINDASH
		if CV_FindVar("k_lightspeeddash").string == "On"
			player.mo.colorized = true
			S_StopSoundByID(player.mo, sfx_salspd)
			S_StartSound(player.mo, sfx_sready)
			player.lightspeeddashable = 1
			player.mo.state = S_PLAY_STND
		end
	end
	if CV_FindVar("k_lightspeeddash").string == "Off" and player.fcolorize == true
		player.fcolorize = false
	end
end)


addHook("PlayerThink", function(player)
	if (player.mo.state == S_PLAY_SPINDASH)
	and player.mo.skin == "kombi"
	and not (player.cmd.buttons & BT_SPIN)
	and not (player.pflags & PF_SPINNING)
		if player.lightspeeddashable > 0
			if CV_FindVar("k_lightspeeddash").string == "On"
				return
			end
		end
		player.mo.state = S_PLAY_ROLL
		player.pflags = $|PF_SPINNING
		P_Thrust(player.mo, player.drawangle, 55*FRACUNIT)
		S_StopSoundByID(player.mo, sfx_salspd)
		S_StartSound(player.mo, sfx_zoom)
	end
end)


-- Others


addHook("PlayerThink", function(player)
	if (player.mo.skin == "kombi") and (player.mo.state == S_PLAY_SPINDASH) and not srb2p
		player.mo.momz = 0
        player.mo.momy = 0
		player.mo.momx = 0
	end	
end)


addHook("PlayerCanDamage", function(player)
    if (player.mo.skin == "kombi") and not srb2p
    and (player.mo.state == S_PLAY_SPINDASH)
        return true
    end
end)

-- Commence Lightspeed Dash Section

local lspdrange = 360*FRACUNIT
local lspdspeed = 60*FRACUNIT
local lspdskin = "kombi"

local function P_RingNearby(player) -- Is a ring in range?
    local closest
	closest = NULL

	searchBlockmap("objects", function(refmobj, mo)

		if mo.health <= 0 then -- Not a valid ring
			return
        end

		if not mo.type == 107 then
			return
        end

		if P_AproxDistance(P_AproxDistance(player.mo.x-mo.x, player.mo.y-mo.y),
			player.mo.z-mo.z) > lspdrange then -- Out of range
			return
        end

		if not P_CheckSight(player.mo, mo) then -- Out of sight
			return
        end

		if closest and P_AproxDistance(P_AproxDistance(player.mo.x-mo.x, player.mo.y-mo.y),
			player.mo.z-mo.z) > P_AproxDistance(P_AproxDistance(player.mo.x-closest.x,
			player.mo.y-closest.y), player.mo.z-closest.z) then
			return
        end

		-- Found a target
		closest = mo
    end, player.mo, player.mo.x-lspdrange, player.mo.x+lspdrange, player.mo.y-lspdrange, player.mo.y+lspdrange)

	if closest then
		return true
    end

	return false
end

local function P_LightDash(source, enemy) -- Home in on your target

	if not source.tracer then
		return -- Nothing to home in on!
    end

	-- adjust direction
	local dest = source.tracer

	if not dest then
		return
    end
	
	-- change angle
	source.player.angle = R_PointToAngle2(source.x, source.y, enemy.x, enemy.y)

	-- change slope
	local dist = P_AproxDistance(P_AproxDistance(dest.x - source.x, dest.y - source.y), dest.z - source.z);

	if dist < 1 then
		dist = 1
    end

	source.momx = FixedMul(FixedDiv(dest.x - source.x, dist), lspdspeed)
	source.momy = FixedMul(FixedDiv(dest.y - source.y, dist), lspdspeed)
	source.momz = FixedMul(FixedDiv(dest.z - source.z, dist), lspdspeed)
end

local function P_LookForRings(player)

    local found
    player.mo.target = NULL
    player.mo.tracer = NULL

	searchBlockmap("objects", function(refmobj, mo)

		if mo.health <= 0 then -- Not a valid ring
			return
        end

		if not mo.type == 107 then
			return
        end

		if P_AproxDistance(P_AproxDistance(player.mo.x-mo.x, player.mo.y-mo.y),
			player.mo.z-mo.z) > lspdrange then -- Out of range
			return
        end

		if not P_CheckSight(player.mo, mo) then -- Out of sight
			return
        end

		if player.mo.target and P_AproxDistance(P_AproxDistance(player.mo.x-mo.x, player.mo.y-mo.y), player.mo.z-mo.z) > 
			P_AproxDistance(P_AproxDistance(player.mo.x-player.mo.target.x, player.mo.y-player.mo.target.y), player.mo.z-player.mo.target.z) then
			return
        end

		-- Found a target
		if mo.type == 107 then
			found = true
			player.mo.target = mo
			player.mo.tracer = mo
        end
    end, player.mo, player.mo.x-lspdrange, player.mo.x+lspdrange, player.mo.y-lspdrange, player.mo.y+lspdrange)

	if found and player.mo.tracer.type == 107 then
		player.lightdash = TICRATE
		P_ResetPlayer(player)
		player.mo.state = S_PLAY_FALL
		P_ResetScore(player)
		P_LightDash(player.mo, player.mo.tracer)
		return
	else
		player.lightspeeddashable = 0
		player.lightdash = 0
		player.attackdown = false
    end
	player.mo.momx = FixedDiv(player.mo.momx,2*FRACUNIT)
	player.mo.momy = FixedDiv(player.mo.momy,2*FRACUNIT)
end

addHook("PlayerThink", function(player)
	-- aqua: checks i added
	if (lspdskin and player.mo.skin ~= lspdskin)
	or player.exiting or P_PlayerInPain(player)
	-- or player.pflags & PF_GLIDING or player.pflags & PF_SLIDING or player.pflags & PF_SHIELDABILITY
	or (player.climbing or player.powers[pw_carry]) then
		player.lightdash = nil
		return
	end

	-- aqua: i removed the player.lightdashallowed variable for optimization purposes.
		-- because, before, to make that unused variable work, i would need to search for rings every frame,
		-- whilst now, i only have to search for rings when you press the light dash button.
		-- however, if you intend on porting the HUD element for the light dash from final demo, you will need to reimplement it.
	if player.lightspeeddashable >= 1 then
			if not (player.cmd.buttons & BT_SPIN)
				if player.lightspeedhold >= leveltime
					P_InstaThrust(player.mo, player.mo.angle, 60*FRACUNIT)
				end
				if player.lightspeeddashable == 1
					player.mo.colorized = false
					S_StartSound(player.mo, sfx_lspdst)
					S_StartSound(player.mo, sfx_slspgo)
					player.lightspeedhold = leveltime + 7
					player.lightdash = 0
					player.lightspeeddashable = 2
				end
				if player.lightspeedhold == leveltime and not player.attackdown
					player.attackdown = true
					if P_RingNearby(player) then
						player.lightdash = TICRATE
						player.lightspeedhold = leveltime
					else
						player.mo.momy = 0*FRACUNIT
						player.mo.momx = 0*FRACUNIT
						player.lightspeeddashable = 0
						player.attackdown = false
					end
				end
			end
	else
			S_StopSoundByID(player.mo, sfx_lspdst)
			S_StopSoundByID(player.mo, sfx_lspdlo)
			player.attackdown = false
    end

	if not (player.lightdash == nil)
		if player.lightdash > 0 then
			player.lightdash = $ - 1
			P_LookForRings(player)
			P_SpawnGhostMobj(player.mo)
			player.powers[pw_flashing] = 2
			if player.lightspeedhold - leveltime == -127
				S_StopSoundByID(player.mo, sfx_lspdst)
				S_StartSound(player.mo, sfx_lspdlo)
				player.timervar2 = leveltime + 82
			end
		end
		if leveltime == timervar2
			S_StartSound(player.mo, sfx_lspdlo)
			player.timervar2 = leveltime + 82
		end
    end
end)